#include <avr/io.h>
#include <avr/boot.h>

#include "types.h"
#ifdef MODE_LCD
	#include "lcd.h"
#endif
#include "delay.h"
#include "spi.h"



#define MMC_CS_DDR	DDRB
#define MMC_CS_PIN	PINB
#define MMC_CS_PORT	PORTB
#define PIN_MMC_CS PB2

// look in datasheet for complete list of commands
#define MMC_CMD_0_GO_IDLE				0
#define MMC_CMD_1_SEND_OP_COND			1
#define MMC_CMD_9_SEND_CSD				9
#define MMC_CMD_10_SEND_CID				10
#define MMC_CMD_12_STOP					12
#define MMC_CMD_13_SEND_STATUS			13
#define MMC_CMD_16_BLOCKLEN				16
#define MMC_CMD_17_READ_SINGLE			17
#define MMC_CMD_18_READ_MULTIPLE		18
#define MMC_CMD_24_WRITE_SINGLE			0x18

// bit definitions for R1
#define MMC_R1_IN_IDLE 	0x01
#define MMC_R1B_BUSY_BYTE 0x00

// different defines for tokens etc ...
#define MMC_START_TOKEN_SINGLE  0xfe
#define MMC_START_TOKEN_MULTI   0xfc



// starts the read process of a sector
//static void MMC_get_sec_start(void);
// gets the next byte from the MMC card
//inline static u08 MMC_get_sec_next(void);
// stop read process of a sector
//static void MMC_get_sec_stop(void);

// deselcets the MMC
inline static void MMC_CS_deselect(void);
//selects the MMC for transmission
inline static void MMC_CS_select(void);

static void MMC_cleanup(void);


//---------------------------------------



u16 sect, bufpos;

// sends a block of data from the mem over the spi bus
static void spi_io_mem(u08* data, u16 length){
	// transmit 'length' bytes over spi
	while(length){
		spi_io(*data);
		data++;
		length--;
	};
};

// send cmd + arguments + default crc for init command. once in spi mode 
// we dont need crc so we can keep this constant
static void MMC_send_cmd(u16 datah, u16 datal){
	// default command sequence
	u08 buffer[6] ;
	// fill sequence with our specific data
	buffer[0]=0x40 | MMC_CMD_17_READ_SINGLE;
	buffer[1]=0;
	buffer[2]=datah&0xff;
	buffer[3]=(datal&0xff00)>>8;
	buffer[4]=0;
	buffer[5]=0x95;
	// dispach data
	spi_io_mem(buffer,6);
};


/*
// send cmd + arguments + default crc for init command. once in spi mode 
// we dont need crc so we can keep this constant
static void MMC_send_cmd(void)
{
	// default command sequence
	spi_io(0x40 | MMC_CMD_17_READ_SINGLE);
	spi_io(0);
	spi_io(adr/256);
	spi_io(adr);
	adr += 2;
	spi_io(0);
	spi_io(0x95);
};
*/

// gets a 1 byte long R1
static u08 MMC_get_R1(void){
	u08 retval;
	u08 max_errors = 128;
	// wait for first valid response byte
	do{
		retval = spi_io(0xff);
		max_errors--;
	}while(  (retval & 0x80) && (max_errors>0));
	return retval;
};

// waits for the card to send the start data block token
void MMC_wait_for_start_token(u08 max_errors){
  u08 retval;
  do{
    // get a byte from the spi bus
    retval = spi_io(0xff);
    // keep track of the trys
    max_errors--;   
  }while((retval != MMC_START_TOKEN_SINGLE) );
};

/*
#define MMC_CS_select() cbi(MMC_CS_PORT, PIN_MMC_CS)
#define MMC_CS_deselect() sbi(MMC_CS_PORT, PIN_MMC_CS)
*/
// selects the CS for spi xfer
inline static void MMC_CS_select(void){
	// pull down the MMC CS line
	cbi(MMC_CS_PORT, PIN_MMC_CS);
};

// deselects the CS for spi xfer
inline static void MMC_CS_deselect(void){
	// pull up the MMC CS card
	sbi(MMC_CS_PORT, PIN_MMC_CS);
};


// stops the MMC transmission and sends the clock cycles needed by the mmc for cleanup
static void MMC_cleanup(void){
	// deselect the MMC card
	MMC_CS_deselect();
	// pulse the SCK 16 times
	spi_io(0xff);
	spi_io(0xff);
};



// starts the read process of a sector
static void MMC_get_sec_start(u16 sectorh, u16 sectorl){
	spi_io(0xff);
	spi_io(0xff);
	// turn sectors into byte addr
	sectorh = sectorl >> 7;
	sectorl = sectorl << 9;
	// select card
	MMC_CS_select();
	// tell the MMC card that we want to know its status
	MMC_send_cmd(sectorh, sectorl);
	// get the response
	if (MMC_get_R1()) {
		MMC_cleanup();
		return;
	}
	// wait till the mmc starts sending data
	MMC_wait_for_start_token(255);
};



// stop read process of a sector
static void MMC_get_sec_stop(void){
		// get the 2 CRC bytes
		spi_io(0xff);
		spi_io(0xff);
		// give enough time
		MMC_cleanup();
};




//-------------------------------------------------------------
// High Level MMC-Funktionen



/*
inline static void mmc_read_block(u08 *buf, u16 cnt, u16 *bufpos) {
	while (cnt--) {
		if (*bufpos == 512) {
			MMC_get_sec_stop();
			MMC_get_sec_start();
			*bufpos = 0;
		}
		*bufpos++;
		*buf = spi_io(0xff);
		buf++;
	}
}
*/
// Lesen beginnen
inline static void mmc_load_start(u16 s) {
	sect = s;
	MMC_get_sec_start(0, (u16)(s));
	bufpos = 0;
}


// Byte lesen
inline static u08 mmc_fetch_byte(void) {
	if (bufpos == 512) {
		MMC_get_sec_stop();
		mmc_load_start(++sect);
	}
	bufpos++;
	return spi_io(0xff);
}


inline static void mmc_read_block(u08 *buf, u16 cnt) {
	while (cnt--) {
		*buf = mmc_fetch_byte();
		buf++;
	}
}




#define PAGESIZEB 128

#ifdef __AVR_ATmega168__
	#define ROMSIZEB 16384
	#define BOOTSIZEB 1024
#endif

#ifdef __AVR_ATmega88__
	#define ROMSIZEB 8192
	#define BOOTSIZEB 1024
#endif

#ifndef F_CPU
	#define F_CPU 8000000
#endif


unsigned char buf[PAGESIZEB];


union temp {
	u08 byte[2];
	u16 word;
};

void main(void) __attribute__ ((naked));

void main(void)
{
	u16 sector, bufpos, blkcnt = OCR1A<<2, cnt, i;
	cli();

#ifdef MODE_LCD
	lcd_clear();
	delay_ms(5);
	lcd_string(DISP_MAIN, 0);
	delay_ms(100);
#endif
	sector = TCNT1;
	//sector <<= 1;
	//adr = sector;
	if (sector == 0 || blkcnt == 0)
		while(1);
#ifdef MODE_LCD
	lcd_string(DISP_WRITE, 1);
#endif

	//MMC_get_sec_start();
	//bufpos = 0;
	mmc_load_start(TCNT1);
	cnt = 0;

	while (blkcnt) {
#ifdef MODE_LCD
		lcd_setcur(14, 1);
		lcd_hex_u16(cnt);
#endif
		mmc_read_block(buf, PAGESIZEB);//, &bufpos);
		u32 page = cnt<<7;
		boot_page_erase(page);
		boot_spm_busy_wait();

		u08 wbufaddr = 0;
		for (i=0; i<SPM_PAGESIZE; i+=2)
		{
		    // Set up little-endian word.
		    union temp w;
			w.byte[0] = buf[wbufaddr++];
		    w.byte[1] = buf[wbufaddr++];
		    boot_page_fill(page + i, w.word);
		}
		boot_page_write(page);     // Store buffer in flash page.
		boot_spm_busy_wait();       // Wait until the memory is written.

		cnt++;
		blkcnt--;
	}
	//mmc_complete_read(&bufpos);
	MMC_get_sec_stop();
#ifdef MODE_LCD
	lcd_string(DISP_DONE, 1);
	while(1);
#endif
}


